"""Added createdDate to BranchIndex

Peek Plugin Database Migration Script

Revision ID: 361988cd327d
Revises: 8ef78f862ac8
Create Date: 2019-05-07 22:14:27.012735

"""

# revision identifiers, used by Alembic.
revision = '361988cd327d'
down_revision = '8ef78f862ac8'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('BranchIndex', sa.Column('createdDate', sa.DateTime(timezone=True), nullable=False), schema='pl_diagram')
    op.add_column('BranchIndex', sa.Column('updatedDate', sa.DateTime(timezone=True), nullable=True), schema='pl_diagram')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('BranchIndex', 'updatedDate', schema='pl_diagram')
    op.drop_column('BranchIndex', 'createdDate', schema='pl_diagram')
    # ### end Alembic commands ###


import logging
from typing import Callable

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import \
    ACIProcessorQueueControllerABC, ACIProcessorQueueBlockItem
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import \
    ACIProcessorStatusNotifierABC
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import \
    ACIProcessorQueueTupleABC
from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexUpdateHandler import \
    ClientLocationIndexUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.LocationIndex import \
    LocationIndexCompilerQueue, LocationIndexCompiled, LocationIndex

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.locationIndexCompilerQueueStatus = state
        self._adminStatusController.status.locationIndexCompilerQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.locationIndexCompilerProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.locationIndexCompilerLastError = error
        self._adminStatusController.notify()


class LocationCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 2.000

    QUEUE_BLOCKS_MAX = 5
    QUEUE_BLOCKS_MIN = 0

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = LocationIndexCompilerQueue
    _VacuumDeclaratives = (LocationIndexCompilerQueue,
                           LocationIndex, LocationIndexCompiled)

    def __init__(self, dbSessionCreator,
                 statusController: StatusController,
                 clientLocationUpdateHandler: ClientLocationIndexUpdateHandler,
                 readyLambdaFunc: Callable):
        ACIProcessorQueueControllerABC \
            .__init__(self, dbSessionCreator, _Notifier(statusController))
        # Disabled
        # isProcessorEnabledCallable=readyLambdaFunc)

        self._clientLocationUpdateHandler: ClientLocationIndexUpdateHandler \
            = clientLocationUpdateHandler

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_plugin_diagram._private.worker.tasks.LocationIndexCompilerTask import \
            compileLocationIndex

        return compileLocationIndex.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientLocationUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return '''
                 with sq_raw as (
                    SELECT "id", "indexBucket"
                    FROM pl_diagram."LocationIndexCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "indexBucket"
                    FROM sq_raw
                    GROUP BY  "indexBucket"
                    HAVING count("indexBucket") > 1
                )
                DELETE
                FROM pl_diagram."LocationIndexCompilerQueue"
                     USING sq sq1
                WHERE pl_diagram."LocationIndexCompilerQueue"."id" != sq1."minId"
                    AND pl_diagram."LocationIndexCompilerQueue"."id" > %(id)s
                    AND pl_diagram."LocationIndexCompilerQueue"."indexBucket" = sq1."indexBucket"

            ''' % {'id': lastFetchedId, 'limit': dedupeLimit}



import logging
from datetime import datetime
from typing import List

import pytz
from sqlalchemy import select, and_

from peek_plugin_diagram._private.storage.DispIndex import DispIndexerQueue
from peek_plugin_diagram._private.storage.LiveDbDispLink import LiveDbDispLink
from peek_plugin_diagram._private.storage.ModelSet import ModelSet, ModelCoordSet
from peek_plugin_diagram._private.worker.tasks.LiveDbDisplayValueConverter import \
    LiveDbDisplayValueConverter
from peek_plugin_livedb.tuples.LiveDbDisplayValueTuple import LiveDbDisplayValueTuple

logger = logging.getLogger(__name__)


class WorkerApiImpl:
    """ Worker Api

    This class allows other classes to work with the Diagram plugin on the
    worker service.

    """
    _FETCH_SIZE = 5000

    @classmethod
    def updateLiveDbDisplayValues(cls,
                                  ormSession,
                                  modelSetKey: str,
                                  liveDbRawValues: List[LiveDbDisplayValueTuple]
                                  ) -> None:
        if not liveDbRawValues:
            return

        startTime = datetime.now(pytz.utc)

        modelSetId = (ormSession.query(ModelSet.id)
                      .filter(ModelSet.key == modelSetKey)
                      .one()
                      .id)

        translater = LiveDbDisplayValueConverter.create(
            ormSession, modelSetId
        )

        for item in liveDbRawValues:
            if item.dataType is None:
                logger.warning("LiveDB key %s is missing dataType", item.key)
                continue

            item.displayValue = translater.translate(item.dataType, item.rawValue)

        logger.debug("Converted %s LiveDB Raw Values in %s",
                     len(liveDbRawValues), (datetime.now(pytz.utc) - startTime))

    @classmethod
    def liveDbDisplayValueUpdateNotify(cls,
                                       ormSession,
                                       modelSetKey: str,
                                       updatedKeys: List[str]):

        logger.debug("TODO TODO - liveDbDisplayValueUpdateNotify coordSetId")
        linkTable = LiveDbDispLink.__table__
        coordSetTable = ModelCoordSet.__table__
        queueTable = DispIndexerQueue.__table__

        modelSetId = ormSession.query(ModelSet.id) \
            .filter(ModelSet.key == modelSetKey) \
            .one() \
            .id

        stmt = select([linkTable.c.dispId]) \
            .select_from(linkTable
                         .join(coordSetTable,
                               linkTable.c.coordSetId == coordSetTable.c.id)) \
            .where(and_(linkTable.c.liveDbKey.in_(updatedKeys),
                        coordSetTable.c.modelSetId == modelSetId))

        ins = queueTable.insert().from_select(['dispId'], stmt)
        ormSession.execute(ins)



import logging
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

from twisted.internet.defer import DeferredList, Deferred, inlineCallbacks
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

from peek_abstract_chunked_index.private.client.handlers.ACICacheHandlerABC import \
    ACICacheHandlerABC
from peek_abstract_chunked_index.private.tuples.ACIUpdateDateTupleABC import \
    ACIUpdateDateTupleABC
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.client.controller.GridCacheController import \
    GridCacheController
from peek_plugin_diagram._private.server.client_handlers.ClientGridLoaderRpc import \
    ClientGridLoaderRpc
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndexCompiled

logger = logging.getLogger(__name__)

clientGridWatchUpdateFromDeviceFilt = {'key': "clientGridWatchUpdateFromDevice"}
clientGridWatchUpdateFromDeviceFilt.update(diagramFilt)

#: This the type of the data that we get when the clients observe new grids.
DeviceGridT = Dict[str, datetime]


# ModelSet HANDLER
class GridCacheHandler(ACICacheHandlerABC):
    _UpdateDateTuple: ACIUpdateDateTupleABC = GridKeyIndexCompiled
    _updateFromServerFilt: Dict = clientGridWatchUpdateFromDeviceFilt
    _logger: logging.Logger = logger

    def __init__(self, cacheController: GridCacheController, clientId: str):
        """ App Grid Handler

        This class handles the custom needs of the desktop/mobile apps observing grids.

        """
        ACICacheHandlerABC.__init__(self, cacheController, clientId)

        # We need to know who is watching what so we can tell the server.
        self._observedGridKeysByVortexUuid = defaultdict(list)
        self._observedVortexUuidsByGridKey = defaultdict(list)

        # We're not using this
        del self._uuidsObserving

    # ---------------
    # Filter out offline vortexes

    def _filterOutOfflineVortexes(self):
        # TODO, Change this to observe offline vortexes
        # This depends on the VortexFactory offline observable implementation.
        # Which is incomplete at this point :-|

        vortexUuids = set(VortexFactory.getRemoteVortexUuids())
        vortexUuidsToRemove = set(self._observedGridKeysByVortexUuid) - vortexUuids

        if not vortexUuidsToRemove:
            return

        for vortexUuid in vortexUuidsToRemove:
            del self._observedGridKeysByVortexUuid[vortexUuid]

        self._rebuildStructs()

    # ---------------
    # Process update from the server

    def notifyOfUpdate(self, gridKeys: List[str]):
        """ Notify of Grid Updates

        This method is called by the client.GridCacheController when it receives updates
        from the server.

        """
        self._filterOutOfflineVortexes()

        payloadsByVortexUuid = defaultdict(Payload)

        for gridKey in gridKeys:

            gridTuple = self._cacheController.encodedChunk(gridKey)
            vortexUuids = self._observedVortexUuidsByGridKey.get(gridKey, [])

            # Queue up the required client notifications
            for vortexUuid in vortexUuids:
                logger.debug("Sending unsolicited grid %s to vortex %s",
                             gridKey, vortexUuid)
                payloadsByVortexUuid[vortexUuid].tuples.append(gridTuple)

        # Send the updates to the clients
        dl = []
        for vortexUuid, payload in list(payloadsByVortexUuid.items()):
            payload.filt = clientGridWatchUpdateFromDeviceFilt

            # Serliase in thread, and then send.
            d = payload.makePayloadEnvelopeDefer()
            d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
            d.addCallback(VortexFactory.sendVortexMsg, destVortexUuid=vortexUuid)
            dl.append(d)

        # Log the errors, otherwise we don't care about them
        dl = DeferredList(dl, fireOnOneErrback=True)
        dl.addErrback(vortexLogFailure, logger, consumeError=True)

    # ---------------
    # Process observes from the devices

    @inlineCallbacks
    def _processObserve(self, payloadEnvelope: PayloadEnvelope,
                        vortexUuid: str,
                        sendResponse: SendVortexMsgResponseCallable,
                        **kwargs):
        cacheAll = payloadEnvelope.filt.get("cacheAll") == True

        payload = yield payloadEnvelope.decodePayloadDefer()

        lastUpdateByGridKey: DeviceGridT = payload.tuples[0]

        if not cacheAll:
            gridKeys = list(lastUpdateByGridKey.keys())
            self._observedGridKeysByVortexUuid[vortexUuid] = gridKeys
            self._rebuildStructs()

        self._replyToObserve(payload.filt,
                             lastUpdateByGridKey,
                             sendResponse,
                             cacheAll=cacheAll)

    def _rebuildStructs(self) -> None:
        """ Rebuild Structs

        Rebuild the reverse index of uuids by grid key.

        :returns: None
        """
        # Rebuild the other reverse lookup
        newDict = defaultdict(list)

        for vortexUuid, gridKeys in self._observedGridKeysByVortexUuid.items():
            for gridKey in gridKeys:
                newDict[gridKey].append(vortexUuid)

        keysChanged = set(self._observedVortexUuidsByGridKey) != set(newDict)

        self._observedVortexUuidsByGridKey = newDict

        # Notify the server that this client service is watching different grids.
        if keysChanged:
            d = ClientGridLoaderRpc.updateClientWatchedGrids(
                clientId=self._clientId,
                gridKeys=list(self._observedVortexUuidsByGridKey)
            )
            d.addErrback(vortexLogFailure, logger, consumeError=False)

    # ---------------
    # Reply to device observe

    def _replyToObserve(self, filt,
                        lastUpdateByGridKey: DeviceGridT,
                        sendResponse: SendVortexMsgResponseCallable,
                        cacheAll=False) -> None:
        """ Reply to Observe

        The client has told us that it's observing a new set of grids, and the lastUpdate
        it has for each of those grids. We will send them the grids that are out of date
        or missing.

        :param filt: The payload filter to respond to.
        :param lastUpdateByGridKey: The dict of gridKey:lastUpdate
        :param sendResponse: The callable provided by the Vortex (handy)
        :returns: None

        """
        gridTuplesToSend = []

        def sendChunk(toSend):
            if not toSend and not cacheAll:
                return

            payload = Payload(filt=filt, tuples=toSend)
            d: Deferred = payload.makePayloadEnvelopeDefer(compressionLevel=2)
            d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
            d.addCallback(sendResponse)
            d.addErrback(vortexLogFailure, logger, consumeError=True)

        # Check and send any updates
        for gridKey, lastUpdate in lastUpdateByGridKey.items():
            # NOTE: lastUpdate can be null.
            gridTuple = self._cacheController.encodedChunk(gridKey)
            if not gridTuple:
                logger.debug("Grid %s is not in the cache" % gridKey)
                continue

            # We are king, If it's it's not our version, it's the wrong version ;-)
            logger.debug("%s, %s,  %s", gridTuple.lastUpdate == lastUpdate,
                         gridTuple.lastUpdate, lastUpdate)
            if gridTuple.lastUpdate == lastUpdate:
                logger.debug("Grid %s matches the cache" % gridKey)
            else:
                gridTuplesToSend.append(gridTuple)
                logger.debug("Sending grid %s from the cache" % gridKey)

            if len(gridTuplesToSend) == 5 and not cacheAll:
                sendChunk(gridTuplesToSend)
                gridTuplesToSend = []

        sendChunk(gridTuplesToSend)



from vortex.handler.TupleActionProcessor import TupleActionProcessor

from peek_plugin_diagram._private.PluginNames import diagramActionProcessorName
from peek_plugin_diagram._private.PluginNames import diagramFilt
from peek_plugin_diagram._private.server.controller.BranchLiveEditController import \
    BranchLiveEditController
from peek_plugin_diagram._private.server.controller.BranchUpdateController import \
    BranchUpdateController
from peek_plugin_diagram._private.tuples.branch.BranchLiveEditTupleAction import \
    BranchLiveEditTupleAction
from peek_plugin_diagram._private.tuples.branch.BranchUpdateTupleAction import \
    BranchUpdateTupleAction
from .controller.StatusController import StatusController


def makeTupleActionProcessorHandler(statusController: StatusController,
                                    branchUpdateController: BranchUpdateController,
                                    branchLiveEditController: BranchLiveEditController):
    processor = TupleActionProcessor(
        tupleActionProcessorName=diagramActionProcessorName,
        additionalFilt=diagramFilt,
        defaultDelegate=statusController)

    processor.setDelegate(BranchUpdateTupleAction.tupleName(), branchUpdateController)

    processor.setDelegate(BranchLiveEditTupleAction.tupleName(), branchLiveEditController)

    return processor



"""Added cascade constraints

Peek Plugin Database Migration Script

Revision ID: 5e2f84d757f2
Revises: cea17a52b9ae
Create Date: 2018-05-03 16:49:27.699014

"""

# revision identifiers, used by Alembic.
revision = '5e2f84d757f2'
down_revision = 'cea17a52b9ae'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('GridKeyIndex_coordSetId_fkey', 'GridKeyIndex', schema='pl_diagram', type_='foreignkey')
    op.create_foreign_key('GridKeyIndex_coordSetId_fkey', 'GridKeyIndex', 'ModelCoordSet', ['coordSetId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram', ondelete='CASCADE')
    op.drop_constraint('LiveDbDispLink_coordSetId_fkey', 'LiveDbDispLink', schema='pl_diagram', type_='foreignkey')
    op.create_foreign_key('LiveDbDispLink_coordSetId_fkey', 'LiveDbDispLink', 'ModelCoordSet', ['coordSetId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram', ondelete='CASCADE')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint('LiveDbDispLink_coordSetId_fkey', 'LiveDbDispLink', schema='pl_diagram', type_='foreignkey')
    op.create_foreign_key('LiveDbDispLink_coordSetId_fkey', 'LiveDbDispLink', 'ModelCoordSet', ['coordSetId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram')
    op.drop_constraint('GridKeyIndex_coordSetId_fkey', 'GridKeyIndex', schema='pl_diagram', type_='foreignkey')
    op.create_foreign_key('GridKeyIndex_coordSetId_fkey', 'GridKeyIndex', 'ModelCoordSet', ['coordSetId'], ['id'], source_schema='pl_diagram', referent_schema='pl_diagram')
    # ### end Alembic commands ###


"""Added overlay flag

Peek Plugin Database Migration Script

Revision ID: 3b41dec74e46
Revises: 97990b4ca6a1
Create Date: 2019-08-12 11:04:53.878245

"""

# revision identifiers, used by Alembic.
revision = '3b41dec74e46'
down_revision = '97990b4ca6a1'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.add_column('DispBase', sa.Column('overlay', sa.Boolean(), server_default='false', nullable=True), schema='pl_diagram')

    op.execute('''
        UPDATE pl_diagram."DispBase"
        SET overlay=sq.val
        FROM (
                SELECT id, "importGroupHash" is null or "importGroupHash" like '%.ov' as val
                FROM pl_diagram."DispBase"
        ) AS sq
        WHERE pl_diagram."DispBase".id=sq.id
    ''')

    op.alter_column('DispBase', 'overlay', type_=sa.Boolean(), nullable=False, schema='pl_diagram')

    op.execute('''
        INSERT INTO pl_diagram."DispCompilerQueue"("dispId")
        SELECT id FROM pl_diagram."DispBase"
    ''')

def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('DispBase', 'overlay', schema='pl_diagram')
    # ### end Alembic commands ###


import logging
from typing import Union

from sqlalchemy.orm import joinedload
from twisted.internet.defer import Deferred

from peek_plugin_diagram._private.storage.Display import DispLevel
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.Tuple import TUPLE_TYPES_BY_NAME
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)

class ServerLookupTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:
        session = self._ormSessionCreator()
        try:
            Lookup = TUPLE_TYPES_BY_NAME[tupleSelector.name]

            if Lookup == DispLevel:
                all = (session.query(DispLevel)
                       .options(joinedload(DispLevel.coordSet)
                                .joinedload(ModelCoordSet.modelSet))
                       .all())

                for item in all:
                    item.data = {"modelSetKey": item.coordSet.modelSet.key}

            else:
                all = (session.query(Lookup)
                       .options(joinedload(Lookup.modelSet))
                       .all())

                for item in all:
                    item.data = {"modelSetKey": item.modelSet.key}

            return Payload(filt, tuples=all).makePayloadEnvelope().toVortexMsg()

        finally:
            session.close()



import hashlib
import logging
from _collections import defaultdict
from base64 import b64encode
from datetime import datetime
from typing import List, Dict

import pytz
from txcelery.defer import DeferrableTask
from vortex.Payload import Payload

from peek_plugin_base.storage.StorageUtil import makeCoreValuesSubqueryCondition, \
    makeOrmValuesSubqueryCondition
from peek_plugin_base.worker import CeleryDbConn
from peek_plugin_base.worker.CeleryApp import celeryApp
from peek_plugin_diagram._private.storage.Display import DispBase
from peek_plugin_diagram._private.storage.LocationIndex import LocationIndexCompiled, \
    LocationIndex
from peek_plugin_diagram._private.storage.LocationIndex import LocationIndexCompilerQueue
from peek_plugin_diagram._private.storage.ModelSet import ModelSet
from peek_plugin_diagram._private.tuples.location_index.LocationIndexTuple import \
    LocationIndexTuple

logger = logging.getLogger(__name__)

""" Location Index Compiler

Compile the location indexes

1) Query for queue
2) Process queue
3) Delete from queue
"""


@DeferrableTask
@celeryApp.task(bind=True)
def compileLocationIndex(self, payloadEncodedArgs: bytes) -> List[str]:
    """ Compile Location Index Task

    :param self: A celery reference to this task
    :param payloadEncodedArgs: An encoded payload containing the queue tuples.
    :returns: A list of grid keys that have been updated.
    """
    argData = Payload().fromEncodedPayload(payloadEncodedArgs).tuples
    queueItems = argData[0]
    queueItemIds: List[int] = argData[1]

    indexBuckets = list(set([i.indexBucket for i in queueItems]))
    modelSetIdByIndexBucket = {i.indexBucket: i.modelSetId for i in queueItems}

    queueTable = LocationIndexCompilerQueue.__table__
    compiledTable = LocationIndexCompiled.__table__
    lastUpdate = datetime.now(pytz.utc).isoformat()

    startTime = datetime.now(pytz.utc)

    session = CeleryDbConn.getDbSession()
    engine = CeleryDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:

        logger.debug("Staring compile of %s queueItems in %s",
                     len(queueItems), (datetime.now(pytz.utc) - startTime))

        # Get Model Sets

        modelSetIds = list(set(modelSetIdByIndexBucket.values()))
        modelSetQry = (
            session.query(ModelSet.key, ModelSet.id)
                .filter(ModelSet.id.in_(modelSetIds))
        )

        modelSetKeyByModelSetId = {o.id: o.key for o in modelSetQry}

        total = 0
        dispData = _buildIndex(session, indexBuckets)

        conn.execute(compiledTable.delete(
            makeCoreValuesSubqueryCondition(engine, compiledTable.c.indexBucket,
                                            indexBuckets)
        ))
        transaction.commit()
        transaction = conn.begin()

        inserts = []
        for indexBucket, jsonStr in dispData.items():
            modelSetId = modelSetIdByIndexBucket[indexBucket]
            modelSetKey = modelSetKeyByModelSetId[modelSetId]

            m = hashlib.sha256()
            m.update(modelSetKey.encode())
            m.update(jsonStr.encode())
            dataHash = b64encode(m.digest()).decode()

            locationIndexTuple = LocationIndexTuple(
                modelSetKey=modelSetKey,
                indexBucket=indexBucket,
                jsonStr=jsonStr,
                lastUpdate=dataHash

            )

            blobData = Payload(tuples=[locationIndexTuple]).toEncodedPayload()

            inserts.append(dict(modelSetId=modelSetId,
                                indexBucket=indexBucket,
                                lastUpdate=dataHash,
                                blobData=blobData))

        if inserts:
            conn.execute(compiledTable.insert(), inserts)

        logger.debug("Compiled %s LocationIndexes, %s missing, in %s",
                     len(inserts),
                     len(indexBuckets) - len(inserts),
                     (datetime.now(pytz.utc) - startTime))

        total += len(inserts)

        conn.execute(queueTable.delete(
            makeCoreValuesSubqueryCondition(engine, queueTable.c.id, queueItemIds)
        ))

        transaction.commit()
        logger.info("Compiled and Comitted %s LocationIndexCompileds in %s",
                    total, (datetime.now(pytz.utc) - startTime))

        return indexBuckets

    except Exception as e:
        transaction.rollback()
        # logger.warning(e)  # Just a warning, it will retry
        logger.exception(e)
        raise self.retry(exc=e, countdown=2)

    finally:
        conn.close()
        session.close()


def _buildIndex(session, indexBuckets) -> Dict[str, str]:
    indexQry = (
        session.query(LocationIndex.indexBucket,
                      DispBase.id, DispBase.key, DispBase.locationJson)
            .join(DispBase, DispBase.id == LocationIndex.dispId)
            .filter(makeOrmValuesSubqueryCondition(
            session, LocationIndex.indexBucket, indexBuckets
        ))
            .order_by(DispBase.key, DispBase.id)
    )

    jsonByIndexBucket = {}

    # Create the IndexBucket -> Key -> [Locations] structure
    locationByKeyByBucket = defaultdict(lambda: defaultdict(list))
    for item in indexQry:
        locationByKeyByBucket[item.indexBucket][item.key].append(item.locationJson)

    # Sort each bucket by the key
    for indexBucket, locationByKey in locationByKeyByBucket.items():

        # Create a list of of key, [locationJson, locationJson, locationJson]
        sortedKeyLocations = list(sorted(locationByKey.items(), key=lambda i: i[0]))

        # [even] is a key, [odd] is the locations json string
        indexStructure = []
        for key, locationsJson in sortedKeyLocations:
            # Combine the key and locations into one json array
            indexStructure.append('["%s",' % key + ','.join(locationsJson) + ']')

        # Create the blob data for this index.
        # It will be searched by a binary sort
        jsonByIndexBucket[indexBucket] = '[' + ','.join(indexStructure) + ']'

    return jsonByIndexBucket



"""enabled nulls for text

Peek Plugin Database Migration Script

Revision ID: 4925c894757b
Revises: e9d80c4c087d
Create Date: 2017-11-04 14:43:04.703237

"""

# revision identifiers, used by Alembic.
revision = '4925c894757b'
down_revision = 'e9d80c4c087d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.alter_column('DispText', 'text', type_=sa.String(), nullable=True, schema='pl_diagram')


def downgrade():
    op.alter_column('DispText', 'text', type_=sa.String(length=1000), nullable=False, schema='pl_diagram')



"""Encoded Grids are now GridTuples

Peek Plugin Database Migration Script

Revision ID: 298b291aac2d
Revises: c88ba8f6761f
Create Date: 2017-12-16 19:30:54.032544

"""

# revision identifiers, used by Alembic.
revision = '298b291aac2d'
down_revision = 'c88ba8f6761f'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    op.execute(' TRUNCATE TABLE pl_diagram."GridKeyIndexCompiled" ')
    op.alter_column('GridKeyIndexCompiled', 'blobData', type_=sa.LargeBinary, new_column_name="encodedGridTuple",
                    schema='pl_diagram')
    op.execute('''INSERT INTO pl_diagram."GridKeyCompilerQueue" ("gridKey", "coordSetId")
                SELECT DISTINCT "gridKey", "coordSetId"
                FROM pl_diagram."GridKeyIndex"
                ''')


def downgrade():
    raise Exception("Downgrade is not implemented")




from vortex.handler.TupleDataObservableProxyHandler import TupleDataObservableProxyHandler

from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import \
    BranchIndexCacheController
from peek_plugin_diagram._private.client.controller.CoordSetCacheController import \
    CoordSetCacheController
from peek_plugin_diagram._private.client.controller.GridCacheController import \
    GridCacheController
from peek_plugin_diagram._private.client.controller.LocationIndexCacheController import \
    LocationIndexCacheController
from peek_plugin_diagram._private.client.controller.LookupCacheController import \
    LookupCacheController
from peek_plugin_diagram._private.client.controller.ModelSetCacheController import \
    ModelSetCacheController
from peek_plugin_diagram._private.client.tuple_providers.BranchIndexUpdateDateTupleProvider import \
    BranchIndexUpdateDateTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.BranchTupleProvider import \
    BranchTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientCoordSetTupleProvider import \
    ClientCoordSetTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientDispKeyLocationTupleProvider import \
    ClientDispKeyLocationTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientGroupDispsTupleProvider import \
    ClientGroupDispsTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientLocationIndexUpdateDateTupleProvider import \
    ClientLocationIndexUpdateDateTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientLookupTupleProvider import \
    ClientLookupTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.ClientModelSetTupleProvider import \
    ClientModelSetTupleProvider
from peek_plugin_diagram._private.client.tuple_providers.GridCacheIndexTupleProvider import \
    GridCacheIndexTupleProvider
from peek_plugin_diagram._private.storage.Display import DispLevel, DispTextStyle, \
    DispLineStyle, DispColor, DispLayer
from peek_plugin_diagram._private.storage.ModelSet import ModelCoordSet, ModelSet
from peek_plugin_diagram._private.tuples.GroupDispsTuple import GroupDispsTuple
from peek_plugin_diagram._private.tuples.branch.BranchIndexUpdateDateTuple import \
    BranchIndexUpdateDateTuple
from peek_plugin_diagram._private.tuples.branch.BranchTuple import \
    BranchTuple
from peek_plugin_diagram._private.tuples.grid.GridUpdateDateTuple import \
    GridUpdateDateTuple
from peek_plugin_diagram._private.tuples.location_index.DispKeyLocationTuple import \
    DispKeyLocationTuple
from peek_plugin_diagram._private.tuples.location_index.LocationIndexUpdateDateTuple import \
    LocationIndexUpdateDateTuple


def makeClientTupleDataObservableHandler(tupleObservable: TupleDataObservableProxyHandler,
                                         modelSetCacheController: ModelSetCacheController,
                                         coordSetCacheController: CoordSetCacheController,
                                         gridCacheController: GridCacheController,
                                         lookupCacheController: LookupCacheController,
                                         locationCacheController: LocationIndexCacheController,
                                         branchCacheController: BranchIndexCacheController):
    """" Make CLIENT Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param branchCacheController:
    :param modelSetCacheController:
    :param locationCacheController:
    :param tupleObservable: The tuple observable proxy
    :param coordSetCacheController: The clients coord set cache controller
    :param gridCacheController: The cache controller for the grid
    :param lookupCacheController: The clients lookup cache controller
    :return: An instance of :code:`TupleDataObservableHandler`

    """
    # Register TupleProviders here
    lookupTupleProvider = ClientLookupTupleProvider(lookupCacheController)

    # Add the lookup providers
    tupleObservable.addTupleProvider(DispLevel.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLayer.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispColor.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispLineStyle.tupleName(), lookupTupleProvider)
    tupleObservable.addTupleProvider(DispTextStyle.tupleName(), lookupTupleProvider)

    # Add the CoordSet providers

    tupleObservable.addTupleProvider(ModelSet.tupleName(),
                                     ClientModelSetTupleProvider(modelSetCacheController))

    tupleObservable.addTupleProvider(ModelCoordSet.tupleName(),
                                     ClientCoordSetTupleProvider(coordSetCacheController))

    tupleObservable.addTupleProvider(GridUpdateDateTuple.tupleName(),
                                     GridCacheIndexTupleProvider(gridCacheController))

    tupleObservable.addTupleProvider(
        DispKeyLocationTuple.tupleName(),
        ClientDispKeyLocationTupleProvider(locationCacheController,
                                           coordSetCacheController))

    tupleObservable.addTupleProvider(
        LocationIndexUpdateDateTuple.tupleName(),
        ClientLocationIndexUpdateDateTupleProvider(locationCacheController))

    tupleObservable.addTupleProvider(BranchTuple.tupleName(),
                                     BranchTupleProvider(branchCacheController))

    tupleObservable.addTupleProvider(BranchIndexUpdateDateTuple.tupleName(),
                                     BranchIndexUpdateDateTupleProvider(
                                         branchCacheController))

    tupleObservable.addTupleProvider(
        GroupDispsTuple.tupleName(),
        ClientGroupDispsTupleProvider(gridCacheController)
    )

    return tupleObservable



import logging

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import \
    ACIProcessorQueueControllerABC, ACIProcessorQueueBlockItem
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import \
    ACIProcessorStatusNotifierABC
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import \
    ACIProcessorQueueTupleABC
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkUpdateHandler import \
    BranchIndexChunkUpdateHandler
from peek_plugin_diagram._private.server.controller.StatusController import \
    StatusController
from peek_plugin_diagram._private.storage.branch.BranchIndex import BranchIndex
from peek_plugin_diagram._private.storage.branch.BranchIndexCompilerQueue import \
    BranchIndexCompilerQueue
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import \
    BranchIndexEncodedChunk

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.branchIndexCompilerQueueStatus = state
        self._adminStatusController.status.branchIndexCompilerQueueSize = queueSize
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.branchIndexCompilerProcessedTotal += delta
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.branchIndexCompilerLastError = error
        self._adminStatusController.notify()


class BranchIndexCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 1.000

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = BranchIndexCompilerQueue
    _VacuumDeclaratives = (BranchIndexCompilerQueue,
                           BranchIndex, BranchIndexEncodedChunk)

    def __init__(self, dbSessionCreator,
                 statusController: StatusController,
                 clientUpdateHandler: BranchIndexChunkUpdateHandler):
        ACIProcessorQueueControllerABC.__init__(self, dbSessionCreator,
                                                _Notifier(statusController))

        self._clientUpdateHandler: BranchIndexChunkUpdateHandler = clientUpdateHandler

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_plugin_diagram._private.worker.tasks.branch.BranchIndexCompiler import \
            compileBranchIndexChunk

        return compileBranchIndexChunk.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return '''
                 with sq_raw as (
                    SELECT "id", "chunkKey"
                    FROM pl_diagram."BranchIndexCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "chunkKey"
                    FROM sq_raw
                    GROUP BY  "chunkKey"
                    HAVING count("chunkKey") > 1
                )
                DELETE
                FROM pl_diagram."BranchIndexCompilerQueue"
                     USING sq sq1
                WHERE pl_diagram."BranchIndexCompilerQueue"."id" != sq1."minId"
                    AND pl_diagram."BranchIndexCompilerQueue"."id" > %(id)s
                    AND pl_diagram."BranchIndexCompilerQueue"."chunkKey" = sq1."chunkKey"

            ''' % {'id': lastFetchedId, 'limit': dedupeLimit}



import logging
from typing import Dict

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkUpdateHandlerABC import \
    ACIChunkUpdateHandlerABC
from peek_abstract_chunked_index.private.tuples.ACIEncodedChunkTupleABC import \
    ACIEncodedChunkTupleABC
from peek_plugin_diagram._private.client.controller.GridCacheController import \
    clientGridUpdateFromServerFilt
from peek_plugin_diagram._private.storage.GridKeyIndex import GridKeyIndexCompiled

logger = logging.getLogger(__name__)


class ClientGridUpdateHandler(ACIChunkUpdateHandlerABC):
    _ChunkedTuple: ACIEncodedChunkTupleABC = GridKeyIndexCompiled
    _updateFromServerFilt: Dict = clientGridUpdateFromServerFilt
    _logger: logging.Logger = logger



""" Updates for DispGroup

Peek Plugin Database Migration Script

Revision ID: 8ef78f862ac8
Revises: 1b17ef0cca17
Create Date: 2019-02-17 13:45:55.038615

"""

# revision identifiers, used by Alembic.
revision = '8ef78f862ac8'
down_revision = '1b17ef0cca17'
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column('DispGroup',
                  sa.Column('compileAsTemplate', sa.Boolean(), server_default='false',
                            nullable=False), schema='pl_diagram')

    op.drop_constraint('DispGroupPointer_groupId_fkey', 'DispGroupPointer',
                       schema='pl_diagram', type_='foreignkey')
    op.drop_index('idxDispGroupPointer_groupId', table_name='DispGroupPointer',
                  schema='pl_diagram')
    op.alter_column('DispGroupPointer', 'groupId', new_column_name='targetDispGroupId',
                    type_=sa.INTEGER(), nullable=True,
                    schema='pl_diagram')
    op.create_index('idxDispGroupPointer_targetDispGroupId', 'DispGroupPointer',
                    ['targetDispGroupId'], unique=False, schema='pl_diagram')
    op.create_foreign_key(None, 'DispGroupPointer', 'DispGroup', ['targetDispGroupId'],
                          ['id'], source_schema='pl_diagram',
                          referent_schema='pl_diagram', ondelete='SET NULL')

    op.add_column('DispBase', sa.Column('groupId', sa.Integer(), nullable=True),
                  schema='pl_diagram')
    op.create_index('idx_Disp_groupId', 'DispBase', ['groupId'], unique=False, schema='pl_diagram')
    op.create_foreign_key(None, 'DispBase', 'DispBase', ['groupId'], ['id'],
                          source_schema='pl_diagram', referent_schema='pl_diagram',
                          ondelete='CASCADE')

    op.drop_table('DispGroupItem', schema='pl_diagram')


def downgrade():
    raise NotImplementedError()



import logging

from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger

from peek_plugin_base.server.PluginServerEntryHookABC import PluginServerEntryHookABC
from peek_plugin_base.server.PluginServerStorageEntryHookABC import \
    PluginServerStorageEntryHookABC
from peek_plugin_base.server.PluginServerWorkerEntryHookABC import \
    PluginServerWorkerEntryHookABC
from peek_plugin_diagram._private.server.api.DiagramApi import DiagramApi
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkLoadRpc import \
    BranchIndexChunkLoadRpc
from peek_plugin_diagram._private.server.client_handlers.BranchIndexChunkUpdateHandler import \
    BranchIndexChunkUpdateHandler
from peek_plugin_diagram._private.server.client_handlers.ClientGridLoaderRpc import \
    ClientGridLoaderRpc
from peek_plugin_diagram._private.server.client_handlers.ClientGridUpdateHandler import \
    ClientGridUpdateHandler
from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexLoaderRpc import \
    ClientLocationIndexLoaderRpc
from peek_plugin_diagram._private.server.client_handlers.ClientLocationIndexUpdateHandler import \
    ClientLocationIndexUpdateHandler
from peek_plugin_diagram._private.server.controller.BranchIndexCompilerQueueController import \
    BranchIndexCompilerQueueController
from peek_plugin_diagram._private.server.controller.BranchLiveEditController import \
    BranchLiveEditController
from peek_plugin_diagram._private.server.controller.BranchUpdateController import \
    BranchUpdateController
from peek_plugin_diagram._private.server.controller.DispCompilerQueueController import \
    DispCompilerQueueController
from peek_plugin_diagram._private.server.controller.DispImportController import \
    DispImportController
from peek_plugin_diagram._private.server.controller.GridKeyCompilerQueueController import \
    GridKeyCompilerQueueController
from peek_plugin_diagram._private.server.controller.LiveDbWatchController import \
    LiveDbWatchController
from peek_plugin_diagram._private.server.controller.LocationCompilerQueueController import \
    LocationCompilerQueueController
from peek_plugin_diagram._private.server.controller.LookupImportController import \
    LookupImportController
from peek_plugin_diagram._private.storage import DeclarativeBase
from peek_plugin_diagram._private.storage.DeclarativeBase import loadStorageTuples
from peek_plugin_diagram._private.storage.Setting import globalSetting, \
    DISP_COMPILER_ENABLED, LOCATION_COMPILER_ENABLED, BRANCH_COMPILER_ENABLED, \
    GRID_COMPILER_ENABLED, globalProperties
from peek_plugin_diagram._private.tuples import loadPrivateTuples
from peek_plugin_diagram.tuples import loadPublicTuples
from peek_plugin_livedb.server.LiveDBApiABC import LiveDBApiABC
from .TupleActionProcessor import makeTupleActionProcessorHandler
from .TupleDataObservable import makeTupleDataObservableHandler
from .admin_handlers import makeAdminBackendHandlers
from .controller.StatusController import StatusController

logger = logging.getLogger(__name__)


class ServerEntryHook(PluginServerEntryHookABC,
                      PluginServerStorageEntryHookABC,
                      PluginServerWorkerEntryHookABC):
    def __init__(self, *args, **kwargs):
        """" Constructor """
        # Call the base classes constructor
        PluginServerEntryHookABC.__init__(self, *args, **kwargs)

        #: Loaded Objects, This is a list of all objects created when we start
        self._loadedObjects = []

        self._api = None

    def load(self) -> None:
        """ Load

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialiastion steps here.

        """
        loadStorageTuples()
        loadPrivateTuples()
        loadPublicTuples()
        logger.debug("Loaded")

    @property
    def dbMetadata(self):
        return DeclarativeBase.metadata

    @inlineCallbacks
    def start(self):
        """ Start

        This will be called when the plugin is loaded, just after the db is migrated.
        Place any custom initialisation steps here.

        """

        # ----------------
        # Get a reference to the LiveDB API
        liveDbApi: LiveDBApiABC = self.platform.getOtherPluginApi("peek_plugin_livedb")

        # ----------------
        # create the client grid updater
        clientGridUpdateHandler = ClientGridUpdateHandler(self.dbSessionCreator)
        self._loadedObjects.append(clientGridUpdateHandler)

        # ----------------
        # create the client disp key index updater
        clientDispIndexUpdateHandler = ClientLocationIndexUpdateHandler(
            self.dbSessionCreator
        )
        self._loadedObjects.append(clientDispIndexUpdateHandler)

        # ----------------
        # Create the client branch index handler
        clientBranchIndexChunkUpdateHandler = BranchIndexChunkUpdateHandler(
            self.dbSessionCreator
        )
        self._loadedObjects.append(clientBranchIndexChunkUpdateHandler)

        # ----------------
        # create the Status Controller
        statusController = StatusController()
        self._loadedObjects.append(statusController)

        # ----------------
        # Create the GRID KEY queue
        gridKeyCompilerQueueController = GridKeyCompilerQueueController(
            self.dbSessionCreator, statusController, clientGridUpdateHandler
        )
        self._loadedObjects.append(gridKeyCompilerQueueController)

        def locationsCanBeQueuedFunc() -> bool:
            return (not gridKeyCompilerQueueController.isBusy()
                    and not dispCompilerQueueController.isBusy())

        # ----------------
        # Create the LOCATION INDEX queue
        locationIndexCompilerQueueController = LocationCompilerQueueController(
            self.dbSessionCreator, statusController, clientDispIndexUpdateHandler,
            readyLambdaFunc=locationsCanBeQueuedFunc
        )
        self._loadedObjects.append(locationIndexCompilerQueueController)

        # ----------------
        # Create the DISP queue
        dispCompilerQueueController = DispCompilerQueueController(
            self.dbSessionCreator, statusController
        )
        self._loadedObjects.append(dispCompilerQueueController)

        # ----------------
        # Branch Live Edit Controller

        branchLiveEditController = BranchLiveEditController()
        self._loadedObjects.append(branchLiveEditController)

        # ----------------
        # Branch Index Compiler Controller
        branchIndexCompilerQueueController = BranchIndexCompilerQueueController(
            dbSessionCreator=self.dbSessionCreator,
            statusController=statusController,
            clientUpdateHandler=clientBranchIndexChunkUpdateHandler
        )
        self._loadedObjects.append(branchIndexCompilerQueueController)

        # ----------------
        # Create the Tuple Observer
        tupleObservable = makeTupleDataObservableHandler(
            self.dbSessionCreator, statusController,
            branchLiveEditController
        )
        self._loadedObjects.append(tupleObservable)

        # ----------------
        # Tell the status controller about the Tuple Observable
        statusController.setTupleObservable(tupleObservable)
        branchLiveEditController.setTupleObservable(tupleObservable)

        # ----------------
        # Initialise the handlers for the admin interface
        self._loadedObjects.extend(
            makeAdminBackendHandlers(tupleObservable, self.dbSessionCreator)
        )

        # ----------------
        # Create the display object Import Controller
        dispImportController = DispImportController(liveDbWriteApi=liveDbApi.writeApi)
        self._loadedObjects.append(dispImportController)

        # ----------------
        # Create the import lookup controller
        lookupImportController = LookupImportController(
            dbSessionCreator=self.dbSessionCreator
        )
        self._loadedObjects.append(lookupImportController)

        # ----------------
        # Create the update branch controller
        branchUpdateController = BranchUpdateController(
            liveDbWriteApi=liveDbApi.writeApi,
            tupleObservable=tupleObservable,
            liveEditController=branchLiveEditController,
            dbSessionCreator=self.dbSessionCreator
        )
        self._loadedObjects.append(branchUpdateController)

        # ----------------
        # Create the Watch Grid Controller
        liveDbWatchController = LiveDbWatchController(
            liveDbWriteApi=liveDbApi.writeApi,
            liveDbReadApi=liveDbApi.readApi,
            dbSessionCreator=self.dbSessionCreator
        )
        self._loadedObjects.append(liveDbWatchController)

        # ----------------
        # Create the GRID API for the client
        self._loadedObjects.extend(
            ClientGridLoaderRpc(liveDbWatchController=liveDbWatchController,
                                dbSessionCreator=self.dbSessionCreator)
                .makeHandlers()
        )

        # ----------------
        # Create the Branch Index for the client
        self._loadedObjects.extend(
            BranchIndexChunkLoadRpc(dbSessionCreator=self.dbSessionCreator)
                .makeHandlers()
        )

        # ----------------
        # Create the LOCATION API for the client
        self._loadedObjects.extend(
            ClientLocationIndexLoaderRpc(dbSessionCreator=self.dbSessionCreator)
                .makeHandlers()
        )

        # ----------------
        # Initialise the API object that will be shared with other plugins
        self._api = DiagramApi(
            statusController, dispImportController,
            lookupImportController, branchUpdateController,
            self.dbSessionCreator
        )
        self._loadedObjects.append(self._api)

        # ----------------
        # Create the Action Processor
        self._loadedObjects.append(
            makeTupleActionProcessorHandler(statusController,
                                            branchUpdateController,
                                            branchLiveEditController)
        )

        # ----------------
        # Start the queue controller

        settings = yield self._loadSettings()

        if settings[DISP_COMPILER_ENABLED]:
            dispCompilerQueueController.start()

        if settings[GRID_COMPILER_ENABLED]:
            gridKeyCompilerQueueController.start()

        if settings[BRANCH_COMPILER_ENABLED]:
            locationIndexCompilerQueueController.start()

        if settings[LOCATION_COMPILER_ENABLED]:
            branchIndexCompilerQueueController.start()

        logger.debug("Started")

    def stop(self):
        """ Stop

        This method is called by the platform to tell the peek app to shutdown and stop
        everything it's doing
        """
        # Shutdown and dereference all objects we constructed when we started
        while self._loadedObjects:
            self._loadedObjects.pop().shutdown()

        self._api = None

        logger.debug("Stopped")

    def unload(self):
        """Unload

        This method is called after stop is called, to unload any last resources
        before the PLUGIN is unlinked from the platform

        """
        logger.debug("Unloaded")

    @property
    def publishedServerApi(self) -> object:
        """ Published Server API
    
        :return  class that implements the API that can be used by other Plugins on this
        platform service.
        """
        return self._api

    @deferToThreadWrapWithLogger(logger)
    def _loadSettings(self):
        dbSession = self.dbSessionCreator()
        try:
            return {globalProperties[p.key]: p.value
                    for p in globalSetting(dbSession).propertyObjects}

        finally:
            dbSession.close()


    ###### Implement PluginServerWorkerEntryHookABC


import logging
from typing import Dict

from peek_abstract_chunked_index.private.server.client_handlers.ACIChunkUpdateHandlerABC import \
    ACIChunkUpdateHandlerABC
from peek_abstract_chunked_index.private.tuples.ACIEncodedChunkTupleABC import \
    ACIEncodedChunkTupleABC
from peek_plugin_diagram._private.client.controller.BranchIndexCacheController import \
    clientBranchIndexUpdateFromServerFilt
from peek_plugin_diagram._private.storage.branch.BranchIndexEncodedChunk import \
    BranchIndexEncodedChunk

logger = logging.getLogger(__name__)


class BranchIndexChunkUpdateHandler(ACIChunkUpdateHandlerABC):
    _ChunkedTuple: ACIEncodedChunkTupleABC = BranchIndexEncodedChunk
    _updateFromServerFilt: Dict = clientBranchIndexUpdateFromServerFilt
    _logger: logging.Logger = logger



"""added line templates

Peek Plugin Database Migration Script

Revision ID: ec97d31aebb0
Revises: a73574690197
Create Date: 2019-09-14 20:00:19.116158

"""

# revision identifiers, used by Alembic.
revision = 'ec97d31aebb0'
down_revision = 'a73574690197'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('DispEdgeTemplate',
    sa.Column('id', sa.Integer(), autoincrement=False, nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('lineWidth', sa.Integer(), server_default='2', nullable=False),
    sa.Column('lineColorId', sa.Integer(), nullable=True),
    sa.Column('lineStyleId', sa.Integer(), nullable=True),
    sa.Column('startEndType', sa.Integer(), nullable=True),
    sa.Column('endEndType', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['id'], ['pl_diagram.DispBase.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['lineColorId'], ['pl_diagram.DispColor.id'], ),
    sa.ForeignKeyConstraint(['lineStyleId'], ['pl_diagram.DispLineStyle.id'], ),
    sa.PrimaryKeyConstraint('id'),
    schema='pl_diagram'
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('DispEdgeTemplate', schema='pl_diagram')
    # ### end Alembic commands ###